# ruff: noqa: RUF022
from __future__ import annotations

from .models import (
    CREATE,
    DELETE,
    MANAGE,
    READ,
    WRITE,
    Admin,
    Anonymous,
    Authenticated,
    Creator,
    InheritSecurity,
    Manager,
    Owner,
    Permission,
    PermissionAssignment,
    Reader,
    Role,
    RoleAssignment,
    RoleType,
    SecurityAudit,
    Writer,
)
from .service import SecurityService, security

__all__ = [
    # Permissions
    "CREATE",
    "DELETE",
    "MANAGE",
    "READ",
    "WRITE",
    # Roles
    "Admin",
    "Anonymous",
    "Authenticated",
    "Creator",
    "Manager",
    "Owner",
    "Reader",
    "Writer",
    #
    "InheritSecurity",
    "Permission",
    "PermissionAssignment",
    "Role",
    "RoleAssignment",
    "RoleType",
    "SecurityAudit",
    "SecurityService",
    # The security service instance
    "security",
]



from __future__ import annotations

import click
from flask import current_app
from flask.cli import with_appcontext

from .scheduler import run_scheduler


@click.command()
@with_appcontext
def scheduler() -> None:
    config = dict(current_app.config)
    run_scheduler(config)



""""""

from __future__ import annotations

from flask_login import AnonymousUserMixin
from whoosh.analysis import (
    CharsetFilter,
    LowercaseFilter,
    NgramFilter,
    PathTokenizer,
    RegexTokenizer,
)

#: A Whoosh analyzer that splits on word boundaries and folds accents and case.
from whoosh.fields import (
    DATETIME,
    ID,
    KEYWORD,
    NUMERIC,
    TEXT,
    FieldType,
    Schema,
    SchemaClass,
)
from whoosh.formats import Existence
from whoosh.support.charset import accent_map

from abilian.core.models.subjects import Group, Principal, User
from abilian.core.util import unwrap
from abilian.services.security.models import Anonymous, Role

accent_folder = (
    RegexTokenizer(r"\w+")
    | LowercaseFilter()  # defaults doesn't split on '.'
    | CharsetFilter(accent_map)
)

#: Analyzer for edge-ngrams, from 2 to 6 characters long
edge_ngram = accent_folder | NgramFilter(minsize=2, maxsize=6, at="start")


def EdgeNgramField() -> TEXT:
    return TEXT(analyzer=edge_ngram)


class _DefaultSearchSchema(SchemaClass):
    """General search schema."""

    object_key = ID(stored=True, unique=True)
    id = NUMERIC(bits=64, signed=False, stored=True)
    object_type = ID(stored=True)
    creator = ID(stored=True)
    owner = ID(stored=True)

    #: security index. This list roles and user/group ids allowed to *see* this
    #: content
    allowed_roles_and_users = KEYWORD(stored=True)

    #: tags indexing
    tag_ids = KEYWORD(stored=True)
    tag_text = TEXT(analyzer=accent_folder)

    # hierarchical index of ids path ('/' is the separator)
    parent_ids = FieldType(format=Existence(), analyzer=PathTokenizer(), stored=True)

    name = TEXT(stored=True, analyzer=accent_folder)
    slug = ID(stored=True)
    description = TEXT(stored=True, analyzer=accent_folder)
    text = TEXT(analyzer=accent_folder)


_default_dyn_fields = {
    "*_prefix": EdgeNgramField(),
    "*_at": DATETIME(stored=True, sortable=True),
}


def DefaultSearchSchema(*args, **kwargs) -> Schema:
    schema = _DefaultSearchSchema()
    for name, field in _default_dyn_fields.items():
        schema.add(name, field, glob=True)
    return schema


def indexable_role(role_or_principal: Role | Principal) -> str:
    """Return a string suitable for query against `allowed_roles_and_users`
    field.

    :param role_or_principal: It can be :data:`Anonymous`, :data:`Authenticated`,
      or an instance of :class:`User` or :class:`Group`.
    """
    role_or_principal = unwrap(role_or_principal)

    if isinstance(role_or_principal, AnonymousUserMixin):
        # transform anonymous user to anonymous role
        role_or_principal = Anonymous

    if isinstance(role_or_principal, User) and role_or_principal.is_anonymous:
        # transform anonymous user to anonymous role
        role_or_principal = Anonymous

    if isinstance(role_or_principal, Role):
        return f"role:{role_or_principal.name}"

    if isinstance(role_or_principal, User):
        return f"user:{role_or_principal.id:d}"

    if isinstance(role_or_principal, Group):
        return f"group:{role_or_principal.id:d}"

    raise ValueError(repr(role_or_principal))



"""First cut at a notification system."""

from __future__ import annotations

from flask import current_app as app
from flask import request
from flask_login import current_user
from werkzeug.exceptions import MethodNotAllowed

from abilian.core.extensions import csrf, db
from abilian.core.models.subjects import User
from abilian.i18n import render_template_i18n
from abilian.sbe.apps.communities.security import require_admin
from abilian.sbe.apps.notifications import TOKEN_SERIALIZER_NAME
from abilian.services.auth.views import get_token_status

from ..tasks.social import make_message, send_daily_social_digest_to
from . import notifications

__all__ = ()

route = notifications.route


@require_admin
@route("/debug/social/")
def debug_social():
    """Send a digest to current user, or user with given email.

    Also displays the email in the browser as a result.
    """
    email = request.args.get("email")
    if email:
        user = User.query.filter(User.email == email).one()
    else:
        user = current_user

    msg = make_message(user)

    status = send_daily_social_digest_to(user)

    if status:
        return msg.html
    else:
        return "No message sent."


@route("/unsubscribe_sbe/<token>/", methods=["GET", "POST"])
@csrf.exempt
def unsubscribe_sbe(token: str) -> str:
    expired, invalid, user = get_token_status(token, TOKEN_SERIALIZER_NAME)
    if expired or invalid:
        return render_template_i18n("notifications/invalid-token.html")

    if request.method == "GET":
        return render_template_i18n(
            "notifications/confirm-unsubscribe.html", token=token
        )

    elif request.method == "POST":
        preferences = app.services["preferences"]
        preferences.set_preferences(user, **{"sbe:notifications:daily": False})
        db.session.commit()
        return render_template_i18n("notifications/unsubscribed.html", token=token)
    else:
        raise MethodNotAllowed



"""Singleton for Dramatiq task manager, to permit lazy declaration of actors.
"""

from __future__ import annotations

from flask_dramatiq import Dramatiq

dramatiq = Dramatiq()

# Remove prmeteus from middleware list
# See:
# https://groups.io/g/dramatiq-users/topic/disabling_prometheus/80745532?p=
dramatiq.middleware = dramatiq.middleware[1:]



from __future__ import annotations

import csv
import io
import json
from collections.abc import Sequence
from os.path import splitext
from typing import IO, Any

from flask import current_app, flash, g, redirect, render_template, request, url_for
from validate_email import validate_email
from werkzeug.wrappers import Response

from abilian.core.extensions import db
from abilian.core.models.subjects import User
from abilian.core.signals import activity
from abilian.core.util import unwrap
from abilian.i18n import _
from abilian.services.auth.views import send_reset_password_instructions
from abilian.web import csrf
from abilian.web.action import Endpoint
from abilian.web.nav import BreadcrumbItem

from .views import route, tab


def wizard_extract_data(
    emails: Sequence[str] = (), csv_data: Sequence[dict[str, str]] = ()
) -> tuple[dict[str, Any] | list[User], list[User], list[dict[str, str | None]]]:
    """Filter data and extract existing accounts, existing members and new
    emails."""
    if csv_data:
        existing_account_csv_roles = {user["email"]: user["role"] for user in csv_data}
        emails = [user["email"] for user in csv_data]
        is_csv = True
    else:
        is_csv = False

    emails = [email.strip() for email in emails]

    already_member_emails = [
        member.email for member in g.community.members if member.email in emails
    ]
    not_member_emails = set(emails) - set(already_member_emails)

    existing_members_objects = [
        user for user in g.community.members if user.email in already_member_emails
    ]

    existing_accounts_objects = User.query.filter(
        User.email.in_(not_member_emails)
    ).all()
    existing_account_emails = [user.email for user in existing_accounts_objects]

    emails_without_account = set(not_member_emails) - set(existing_account_emails)

    accounts_list = []
    for user in existing_accounts_objects:
        account = {
            "email": user.email,
            "first_name": user.first_name,
            "last_name": user.last_name,
            "role": existing_account_csv_roles[user.email] if is_csv else "member",
            "status": "existing",
        }
        accounts_list.append(account)

    if is_csv:
        emails_without_account = [
            csv_account
            for csv_account in csv_data
            if csv_account["email"] in emails_without_account
        ]
        existing_accounts_objects = {
            "account_objects": existing_accounts_objects,
            "csv_roles": existing_account_csv_roles,
        }

        for csv_account in emails_without_account:
            account = {}
            account["email"] = csv_account["email"]
            account["first_name"] = csv_account["first_name"]
            account["last_name"] = csv_account["last_name"]
            account["role"] = csv_account["role"]
            account["status"] = "new"
            accounts_list.append(account)

    else:
        for email in emails_without_account:
            account = {}
            account["email"] = email
            account["first_name"] = ""
            account["last_name"] = ""
            account["role"] = "member"
            account["status"] = "new"
            accounts_list.append(account)

    return existing_accounts_objects, existing_members_objects, accounts_list


def wizard_read_csv(csv_file: IO[str]) -> list[dict[str, str]]:
    """Read new members data from CSV file."""

    if hasattr(csv_file, "filename"):
        filename = csv_file.filename
    else:
        filename = csv_file.name
    file_extension = splitext(filename)[-1]
    if file_extension != ".csv":
        return []

    raw_data = csv_file.read()
    if isinstance(raw_data, bytes):
        data = raw_data.decode("utf8")  # FIXME
    else:
        data = raw_data

    contents = csv.reader(io.StringIO(data), delimiter=";")

    new_accounts = []

    for row in contents:
        account = {}
        if len(row) != 4:
            continue

        email = row[0].strip()
        first_name = row[1].strip()
        last_name = row[2].strip()
        role = row[3].strip()

        if not validate_email(email):
            continue
        if role.lower() not in ["manager", "member"]:
            continue

        account["email"] = email
        account["first_name"] = first_name
        account["last_name"] = last_name
        account["role"] = role
        new_accounts.append(account)

    return new_accounts


@route("/<string:community_id>/members/wizard/step1")
@tab("members")
def wizard_data_insertion():
    """Insert new members data into the community via emails or CSV file."""
    g.breadcrumb.append(
        BreadcrumbItem(
            label=_("Members"),
            url=Endpoint("communities.members", community_id=g.community.slug),
        )
    )

    return render_template("community/wizard_add_emails.html")


@route("/<string:community_id>/members/wizard/step2", methods=["GET", "POST"])
@csrf.protect
@tab("members")
def wizard_check_data() -> Response | str:
    """Filter and detect existing members, existing accounts and new emails."""
    if request.method == "GET":
        return redirect(url_for(".members", community_id=g.community.slug))

    g.breadcrumb.append(
        BreadcrumbItem(
            label=_("Members"),
            url=Endpoint("communities.members", community_id=g.community.slug),
        )
    )

    is_csv = False
    if request.form.get("wizard-emails"):
        wizard_emails = request.form["wizard-emails"].split(",")
        (
            existing_accounts_object,
            existing_members_objects,
            final_email_list,
        ) = wizard_extract_data(wizard_emails)
        final_email_list_json = json.dumps(final_email_list)

    else:
        is_csv = True
        accounts_data = wizard_read_csv(request.files["csv_file"])
        if not accounts_data:
            flash(_("To add new members, please follow the CSV file model."), "warning")
            return redirect(
                url_for(".wizard_data_insertion", community_id=g.community.slug)
            )

        (
            existing_accounts,
            existing_members_objects,
            final_email_list,
        ) = wizard_extract_data(csv_data=accounts_data)
        existing_accounts_object = existing_accounts["account_objects"]
        existing_accounts_csv_roles = existing_accounts["csv_roles"]
        final_email_list_json = json.dumps(final_email_list)

    if not final_email_list:
        flash(_("No new members were found"), "warning")
        return redirect(
            url_for(".wizard_data_insertion", community_id=g.community.slug)
        )

    ctx = {
        "existing_accounts_object": existing_accounts_object,
        "csv_roles": existing_accounts_csv_roles if is_csv else False,
        "wizard_emails": final_email_list_json,
        "existing_members_objects": existing_members_objects,
    }
    return render_template("community/wizard_check_members.html", **ctx)


@route("/<string:community_id>/members/wizard/step3", methods=["GET", "POST"])
@csrf.protect
@tab("members")
def wizard_new_accounts() -> str | Response:
    """Complete new emails information."""
    if request.method == "GET":
        return redirect(url_for(".members", community_id=g.community.slug))

    g.breadcrumb.append(
        BreadcrumbItem(
            label=_("Members"),
            url=Endpoint("communities.members", community_id=g.community.slug),
        )
    )

    wizard_emails = request.form["wizard-emails"]
    wizard_accounts = json.loads(wizard_emails)

    wizard_existing_account = {}
    new_accounts = []

    for user in wizard_accounts:
        if user["status"] == "existing":
            wizard_existing_account[user["email"]] = user["role"]

        elif user["status"] == "new":
            new_accounts.append(user)

    existing_account = json.dumps(wizard_existing_account)

    return render_template(
        "community/wizard_new_accounts.html",
        existing_account=existing_account,
        new_accounts=new_accounts,
    )


@route("/<string:community_id>/members/wizard/complete", methods=["POST"])
@csrf.protect
def wizard_saving() -> Response:
    """Automatically add existing accounts to the current community.

    Create accounts for new emails, add them to the community and send
    them a password reset email.
    """
    community = g.community._model
    existing_accounts = request.form["existing_account"]
    existing_accounts = json.loads(existing_accounts)
    new_accounts = request.form.get("new_accounts")
    new_accounts = json.loads(new_accounts)

    if not (existing_accounts or new_accounts):
        flash(_("No new members were found"), "warning")
        return redirect(url_for(".members", community_id=g.community.slug))

    if existing_accounts:
        for email, role in existing_accounts.items():
            user = User.query.filter(User.email == email).first()
            community.set_membership(user, role)

            app = unwrap(current_app)
            activity.send(app, actor=user, verb="join", object=community)

            db.session.commit()

    if new_accounts:
        for account in new_accounts:
            email = account["email"]
            first_name = account["first_name"]
            last_name = account["last_name"]
            role = account["role"]

            user = User(
                email=email, last_name=last_name, first_name=first_name, can_login=True
            )
            db.session.add(user)

            community.set_membership(user, role)
            app = unwrap(current_app)
            activity.send(app, actor=user, verb="join", object=community)
            db.session.commit()

            send_reset_password_instructions(user)

    flash(_("New members added successfully"), "success")
    return redirect(url_for(".members", community_id=community.slug))



from __future__ import annotations

import jinja2

from abilian.sbe.app import Application
from abilian.services.preferences import preferences

from .panels.sbe_notifications import SbeNotificationsPanel


def register_plugin(app: Application) -> None:
    loader = jinja2.PackageLoader("abilian.sbe.apps.preferences")
    app.register_jinja_loaders(loader)
    preferences.register_panel(SbeNotificationsPanel(), app)



"""Dramatiq tasks related to documents."""

from __future__ import annotations

from collections.abc import Iterator
from contextlib import contextmanager
from typing import TYPE_CHECKING

from loguru import logger
from sqlalchemy.orm import Session

from abilian.core.dramatiq.singleton import dramatiq
from abilian.core.extensions import db
from abilian.services import converter, get_service
from abilian.services.conversion import ConversionError, HandlerNotFoundError

if TYPE_CHECKING:
    from .models import Document


@contextmanager
def get_document(
    document_id: int, session: Session | None = None
) -> Iterator[tuple[Session, Document | None]]:
    """Context manager that yields (session, document)."""
    from .models import Document

    logger.debug(
        "get_document() document_id={document_id} session={session}",
        document_id=document_id,
        session=session,
    )

    if session is None:
        doc_session = db.session()
    else:
        doc_session = session

    with doc_session.begin_nested():
        query = doc_session.query(Document)
        document = query.get(document_id)
        logger.debug(
            "get_document() document={document} session={session}",
            document=document,
            session=session,
        )
        yield (doc_session, document)

    # cleanup
    if session is None:
        doc_session.commit()
        doc_session.close()


@dramatiq.actor(max_retries=20, max_backoff=86400000)
def process_document(document_id: int) -> None:
    """Task to process a document.

    max_retries = 20 (Dramatiq default)
    max_backoff = 86400000 , i.e. 1 day
    """
    logger.debug(
        "process_document() actor : document_id={document_id}",
        document_id=document_id,
    )

    with get_document(document_id) as (session, document):
        if document is None:
            return

        # True = Ok, None means no check performed (no antivirus present)
        is_clean = _run_antivirus(document)
        logger.debug(
            "process_document() document_id={document_id} is_clean={is_clean}",
            document_id=document_id,
            is_clean=is_clean,
        )
        if is_clean is False:
            return
    preview_document.send(document_id)
    convert_document_content.send(document_id)


def _run_antivirus(document: Document) -> bool | None:
    antivirus = get_service("antivirus")
    if antivirus and antivirus.running:
        is_clean = antivirus.scan(document.content_blob)
        if "antivirus_task" in document.content_blob.meta:
            del document.content_blob.meta["antivirus_task"]
        return is_clean
    return None


@dramatiq.actor
def antivirus_scan(document_id):
    """Return antivirus.scan() result."""
    with get_document(document_id) as (session, document):
        if document is None:
            return None
        return _run_antivirus(document)


@dramatiq.actor(max_retries=5)
def preview_document(document_id: int) -> None:
    """Compute the document preview images with its default preview size."""

    with get_document(document_id) as (session, document):
        logger.debug(
            "preview_document() document_id={document_id} document={document}",
            document_id=document_id,
            document=document,
        )
        if document is None:
            # deleted after task queued, but before task run
            return

        convert_to_image(document)


@logger.catch(level="ERROR")
def convert_to_image(doc: Document) -> None:
    logger.debug("convert_to_image() document={document}", document=doc)

    try:
        converter.to_image(
            doc.content_digest,
            doc.content,
            doc.content_type,
            0,
            doc.preview_size,
        )
    except ConversionError as e:
        logger.info(
            "Preview failed for document {doc_name}: {error}",
            doc_name=doc.name,
            error=str(e),
        )


@dramatiq.actor(max_retries=5)
def convert_document_content(document_id: int) -> None:
    """Convert document content."""

    logger.debug(
        "convert_document_content() document_id={document_id}",
        document_id=document_id,
    )

    with get_document(document_id) as (session, doc):
        if doc is None:
            # deleted after task queued, but before task run
            return

        convert_to_pdf(doc)
        convert_to_text(doc)
        extract_metadata(doc)


@logger.catch(level="ERROR")
def convert_to_pdf(doc: Document) -> None:
    logger.debug("convert_to_pdf() document={document}", document=doc)

    if doc.content_type == "application/pdf":
        doc.pdf = doc.content
        return
    try:
        doc.pdf = converter.to_pdf(doc.content_digest, doc.content, doc.content_type)
    except (HandlerNotFoundError, ConversionError) as e:
        doc.pdf = b""
        logger.info(
            "Conversion to PDF failed for document {doc_name}: {error}",
            doc_name=doc.name,
            error=str(e),
        )


@logger.catch(level="ERROR")
def convert_to_text(doc: Document) -> None:
    logger.debug("convert_to_text() document={document}", document=doc)

    try:
        doc.text = converter.to_text(doc.content_digest, doc.content, doc.content_type)
    except ConversionError as e:
        doc.text = ""
        logger.info(
            "Conversion to text failed for document {doc_name}: {error}",
            doc_name=doc.name,
            error=str(e),
        )


@logger.catch(level="ERROR")
def extract_metadata(doc: Document) -> None:
    logger.debug("extract_metadata() document={document}", document=doc)

    doc.extra_metadata = {}
    try:
        doc.extra_metadata = converter.get_metadata(
            doc.content_digest, doc.content, doc.content_type
        )
    except ConversionError as e:
        logger.warning(
            "Metadata extraction failed on document {doc_name}: {error}",
            doc_name=doc.name,
            error=str(e),
        )
    except UnicodeDecodeError as e:
        logger.error(
            "Unicode issue on {doc_name}: {error}",
            doc_name=doc.name,
            error=str(e),
        )
    except Exception as e:
        logger.error(
            "Other issue on {doc_name}: {error}",
            doc_name=doc.name,
            error=str(e),
        )

    if doc.text:
        import langid

        doc.language = langid.classify(doc.text)[0]

    doc.page_num = doc.extra_metadata.get("PDF:Pages", 1)



"""Forum views."""

from __future__ import annotations

from collections import Counter
from datetime import date, datetime, timedelta
from itertools import groupby
from urllib.parse import quote

import sqlalchemy as sa
from flask import current_app, flash, g, make_response, render_template, request
from flask_babel import format_date
from flask_login import current_user
from sqlalchemy.orm import joinedload
from werkzeug.exceptions import BadRequest, NotFound

from abilian.core.util import utc_dt
from abilian.i18n import _, _l
from abilian.sbe.apps.communities.security import is_manager
from abilian.services.viewtracker import viewtracker
from abilian.web import url_for, views
from abilian.web.action import ButtonAction, Endpoint
from abilian.web.nav import BreadcrumbItem
from abilian.web.views import default_view

from ..communities.blueprint import CommunityBlueprint
from ..communities.common import activity_time_format, object_viewers
from ..communities.views import default_view_kw
from .forms import PostEditForm, PostForm, ThreadForm
from .models import Post, PostAttachment, Thread
from .tasks import send_post_by_email

# TODO: move to config
MAX_THREADS = 30

forum = CommunityBlueprint(
    "forum", __name__, url_prefix="/forum", template_folder="templates"
)
route = forum.route


def post_kw_view_func(kw, obj, obj_type, obj_id, **kwargs):
    """kwargs for Post default view."""
    kw = default_view_kw(kw, obj.thread, obj_type, obj_id, **kwargs)
    kw["thread_id"] = obj.thread_id
    kw["_anchor"] = f"post_{obj.id:d}"
    return kw


@forum.url_value_preprocessor
def init_forum_values(endpoint, values):
    g.current_tab = "forum"

    g.breadcrumb.append(
        BreadcrumbItem(
            label=_l("Conversations"),
            url=Endpoint("forum.index", community_id=g.community.slug),
        )
    )


def get_nb_viewers(entities):
    if entities:
        views = viewtracker.get_views(entities=entities)
        threads = [
            thread.entity
            for thread in views
            if thread.user in g.community.members
            and thread.user != thread.entity.creator
        ]

        return Counter(threads)


def get_viewed_posts(entities):
    if not entities:
        return None

    views = viewtracker.get_views(entities=entities, user=current_user)
    all_hits = viewtracker.get_hits(views=views)
    nb_viewed_posts = {}
    for view in views:
        related_hits = [hit for hit in all_hits if hit.view_id == view.id]
        entity = view.entity
        if entity in entities:
            cutoff = related_hits[-1].viewed_at
            nb_viewed_posts[entity] = len(
                [post for post in entity.posts if post.created_at > cutoff]
            )

    never_viewed = set(entities) - {view.entity for view in views}
    for entity in never_viewed:
        nb_viewed_posts[entity] = len(entity.posts) - 1

    return nb_viewed_posts


def get_viewed_times(entities):
    if entities:
        views = viewtracker.get_views(entities=entities)
        views = [
            view
            for view in views
            if view.user != view.entity.creator and view.user in g.community.members
        ]

        all_hits = viewtracker.get_hits(views=views)
        views_id = [view.view_id for view in all_hits]

        viewed_times = Counter(views_id)
        entity_viewed_times = {}
        for view in views:
            if view.entity not in entity_viewed_times:
                entity_viewed_times[view.entity] = viewed_times[view.id]
            else:
                entity_viewed_times[view.entity] += viewed_times[view.id]

        return entity_viewed_times


@route("/")
@route("/<string:filter>")
def index(filter=None):
    query = Thread.query.filter(Thread.community_id == g.community.id).order_by(
        Thread.last_post_at.desc()
    )

    threads = query.all()
    has_more = False

    nb_viewed_times = get_viewed_times(threads)
    for thread in threads:
        thread.nb_views = nb_viewed_times.get(thread, 0)

    dt = None
    if filter == "today":
        dt = timedelta(days=1)
    elif filter == "week":
        dt = timedelta(days=7)
    elif filter == "month":
        dt = timedelta(days=31)
    elif filter == "year":
        dt = timedelta(days=365)
    elif filter:
        raise BadRequest

    if dt:
        cutoff_date = datetime.utcnow() - dt
        threads = [thread for thread in threads if thread.created_at > cutoff_date]

    if dt:
        threads = sorted(threads, key=lambda thread: -thread.nb_views)
    else:
        has_more = query.count() > MAX_THREADS
        threads = query.limit(MAX_THREADS).all()

    nb_viewers = get_nb_viewers(threads)
    nb_viewed_posts = get_viewed_posts(threads)

    return render_template(
        "forum/index.html",
        threads=threads,
        has_more=has_more,
        nb_viewers=nb_viewers,
        nb_viewed_posts=nb_viewed_posts,
        nb_viewed_times=nb_viewed_times,
        activity_time_format=activity_time_format,
    )


def group_monthly(entities_list):
    # We're using Python's groupby instead of SA's group_by here
    # because it's easier to support both SQLite and Postgres this way.
    def grouper(entity):
        return entity.created_at.year, entity.created_at.month

    def format_month(year, month):
        month = format_date(date(year, month, 1), "MMMM").capitalize()
        return f"{month} {year}"

    grouped_entities = groupby(entities_list, grouper)
    grouped_entities = [
        (format_month(year, month), list(entities))
        for (year, month), entities in grouped_entities
    ]
    return grouped_entities


@route("/archives/")
def archives():
    all_threads = (
        Thread.query.filter(Thread.community_id == g.community.id)
        .order_by(Thread.created_at.desc())
        .all()
    )

    grouped_threads = group_monthly(all_threads)
    return render_template("forum/archives.html", grouped_threads=grouped_threads)


@route("/attachments/")
def attachments():
    all_threads = (
        Thread.query.filter(Thread.community_id == g.community.id)
        .options(joinedload("posts"))
        .options(joinedload("posts.attachments"))
        .order_by(Thread.created_at.desc())
        .all()
    )

    posts_with_attachments = []
    for thread in all_threads:
        for post in thread.posts:
            if getattr(post, "attachments", None):
                posts_with_attachments.append(post)
    posts_with_attachments.sort(key=lambda post: post.created_at)
    posts_with_attachments.reverse()

    grouped_posts = group_monthly(posts_with_attachments)
    return render_template("forum/attachments.html", grouped_posts=grouped_posts)


class BaseThreadView:
    Model = Thread
    Form = ThreadForm
    pk = "thread_id"
    base_template = "community/_base.html"

    def can_send_by_mail(self):
        return g.community.type == "participative" or is_manager(user=current_user)

    def prepare_args(self, args, kwargs):
        args, kwargs = super().prepare_args(args, kwargs)
        self.send_by_email = False

        if not self.can_send_by_mail() and "send_by_email" in self.form:
            # remove from html form and avoid validation errors
            del self.form["send_by_email"]

        return args, kwargs

    def index_url(self):
        return url_for(".index", community_id=g.community.slug)

    def view_url(self):
        return url_for(self.obj)


class ThreadView(BaseThreadView, views.ObjectView):
    methods = ["GET", "HEAD"]
    Form = PostForm
    template = "forum/thread.html"

    @property
    def template_kwargs(self):
        kw = super().template_kwargs
        kw["thread"] = self.obj
        kw["is_closed"] = self.obj.closed
        kw["is_manager"] = is_manager(user=current_user)
        kw["viewers"] = object_viewers(self.obj)
        kw["views"] = get_viewed_times([self.obj])
        kw["participants"] = {post.creator for post in self.obj.posts}
        kw["activity_time_format"] = activity_time_format
        viewtracker.record_hit(entity=self.obj, user=current_user)
        return kw


thread_view = ThreadView.as_view("thread")
default_view(forum, Post, None, kw_func=post_kw_view_func)(thread_view)
default_view(forum, Thread, "thread_id", kw_func=default_view_kw)(thread_view)

route("/<int:thread_id>/")(thread_view)
route("/<int:thread_id>/attachments")(
    ThreadView.as_view("thread_attachments", template="forum/thread_attachments.html")
)


class ThreadCreate(BaseThreadView, views.ObjectCreate):
    base_template = "community/_forumbase.html"
    template = "forum/thread_create.html"
    POST_BUTTON = ButtonAction(
        "form", "create", btn_class="primary", title=_l("Post this message")
    )

    title = _("New conversation")

    def init_object(self, args, kwargs):
        args, kwargs = super().init_object(args, kwargs)
        self.thread = self.obj
        return args, kwargs

    def before_populate_obj(self):
        del self.form["attachments"]
        self.message_body = self.form.message.data
        del self.form["message"]

        if "send_by_email" in self.form:
            self.send_by_email = (
                self.can_send_by_mail() and self.form.send_by_email.data
            )
            del self.form["send_by_email"]

    def after_populate_obj(self):
        if self.thread.community is None:
            self.thread.community = g.community._model

        self.post = self.thread.create_post(body_html=self.message_body)
        obj_meta = self.post.meta.setdefault("abilian.sbe.forum", {})
        obj_meta["origin"] = "web"
        obj_meta["send_by_email"] = self.send_by_email
        session = sa.orm.object_session(self.thread)
        uploads = current_app.extensions["uploads"]

        for handle in request.form.getlist("attachments"):
            fileobj = uploads.get_file(current_user, handle)
            if fileobj is None:
                continue

            meta = uploads.get_metadata(current_user, handle)
            name = meta.get("filename", handle)
            mimetype = meta.get("mimetype")

            if not isinstance(name, str):
                name = str(name, encoding="utf-8", errors="ignore")

            if not name:
                continue

            attachment = PostAttachment(name=name)
            attachment.post = self.post

            with fileobj.open("rb") as f:
                attachment.set_content(f.read(), mimetype)
            session.add(attachment)

    def commit_success(self):
        if self.send_by_email:
            message = send_post_by_email.send(self.post.id)
            # task = send_post_by_email.delay(self.post.id)
            meta = self.post.meta.setdefault("abilian.sbe.forum", {})
            # note: message_id is of type str
            meta["send_post_by_email_task"] = message.message_id
            self.post.meta.changed()
            session = sa.orm.object_session(self.post)
            session.commit()

    @property
    def activity_target(self):
        return self.thread.community

    def get_form_buttons(self, *args, **kwargs):
        return [self.POST_BUTTON, views.object.CANCEL_BUTTON]


route("/new_thread/")(ThreadCreate.as_view("new_thread", view_endpoint=".thread"))


class ThreadPostCreate(ThreadCreate):
    """Add a new post to a thread."""

    methods = ["POST"]
    Form = PostForm
    Model = Post

    def init_object(self, args, kwargs):
        # we DO want to skip ThreadCreate.init_object. hence super is not based on
        # ThreadPostCreate
        args, kwargs = super().init_object(args, kwargs)
        thread_id = kwargs.pop(self.pk, None)
        self.thread = Thread.query.get(thread_id)
        Thread.query.filter(Thread.id == thread_id).update(
            {Thread.last_post_at: datetime.utcnow()}
        )
        return args, kwargs

    def after_populate_obj(self):
        super().after_populate_obj()
        session = sa.orm.object_session(self.obj)
        session.expunge(self.obj)
        self.obj = self.post


class ThreadViewers(ThreadView):
    template = "forum/thread_viewers.html"


route("/<int:thread_id>/")(
    ThreadPostCreate.as_view("thread_post", view_endpoint=".thread")
)

route("/<int:thread_id>/viewers")(ThreadViewers.as_view("thread_viewers"))


class ThreadDelete(BaseThreadView, views.ObjectDelete):
    methods = ["POST"]
    _message_success = _('Thread "{title}" deleted.')

    def message_success(self):
        return str(self._message_success).format(title=self.obj.title)


route("/<int:thread_id>/delete")(ThreadDelete.as_view("thread_delete"))


class ThreadCloseView(BaseThreadView, views.object.BaseObjectView):
    """Close / Re-open a thread."""

    methods = ["POST"]
    _VALID_ACTIONS = {"close", "reopen"}
    CLOSED_MSG = _l("The thread is now closed for edition and new contributions.")
    REOPENED_MSG = _l("The thread is now re-opened for edition and new contributions.")

    def prepare_args(self, args, kwargs):
        args, kwargs = super().prepare_args(args, kwargs)
        action = kwargs["action"] = request.form.get("action")
        if action not in self._VALID_ACTIONS:
            raise BadRequest(f"Unknown action: {action!r}")

        return args, kwargs

    def post(self, action=None):
        is_closed = action == "close"
        self.obj.closed = is_closed
        sa.orm.object_session(self.obj).commit()

        msg = self.CLOSED_MSG if is_closed else self.REOPENED_MSG
        flash(str(msg))
        return self.redirect(url_for(self.obj))


route("/<int:thread_id>/close")(ThreadCloseView.as_view("thread_close"))


class ThreadPostEdit(BaseThreadView, views.ObjectEdit):
    Form = PostEditForm
    Model = Post
    pk = "object_id"

    def can_send_by_mail(self):
        # post edit: don't notify every time
        return False

    def init_object(self, args, kwargs):
        # we DO want to skip ThreadCreate.init_object. hence super is not based on
        # ThreadPostCreate
        args, kwargs = super().init_object(args, kwargs)
        thread_id = kwargs.pop("thread_id", None)
        self.thread = self.obj.thread
        assert thread_id == self.thread.id
        return args, kwargs

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["message"] = self.obj.body_html
        return kwargs

    def before_populate_obj(self):
        self.message_body = self.form.message.data
        del self.form["message"]
        self.reason = self.form.reason.data

        self.send_by_email = False
        if "send_by_email" in self.form:
            del self.form["send_by_email"]

        self.attachments_to_remove = self.form["attachments"].delete_files_index
        del self.form["attachments"]

    def after_populate_obj(self):
        session = sa.orm.object_session(self.obj)
        uploads = current_app.extensions["uploads"]
        self.obj.body_html = self.message_body
        obj_meta = self.obj.meta.setdefault("abilian.sbe.forum", {})
        history = obj_meta.setdefault("history", [])
        history.append(
            {
                "user_id": current_user.id,
                "user": str(current_user),
                "date": utc_dt(datetime.utcnow()).isoformat(),
                "reason": self.form.reason.data,
            }
        )
        self.obj.meta["abilian.sbe.forum"] = obj_meta  # trigger change for SA

        attachments_to_remove = []
        for idx in self.attachments_to_remove:
            try:
                idx = int(idx)
            except ValueError:
                continue

            if idx > len(self.obj.attachments):
                continue

            attachments_to_remove.append(self.obj.attachments[idx])

        for att in attachments_to_remove:
            session.delete(att)

        for handle in request.form.getlist("attachments"):
            fileobj = uploads.get_file(current_user, handle)
            if fileobj is None:
                continue

            meta = uploads.get_metadata(current_user, handle)
            name = meta.get("filename", handle)
            mimetype = meta.get("mimetype")

            if not isinstance(name, str):
                name = str(name, encoding="utf-8", errors="ignore")

            if not name:
                continue

            attachment = PostAttachment(name=name, post=self.obj)

            with fileobj.open("rb") as f:
                attachment.set_content(f.read(), mimetype)
            session.add(attachment)


route("/<int:thread_id>/<int:object_id>/edit")(ThreadPostEdit.as_view("post_edit"))


def attachment_kw_view_func(kw, obj, obj_type, obj_id, **kwargs):
    post = obj.post
    kw = default_view_kw(kw, post.thread, obj_type, obj_id, **kwargs)
    kw["thread_id"] = post.thread_id
    kw["post_id"] = post.id
    return kw


@route("/<int:thread_id>/posts/<int:post_id>/attachment/<int:attachment_id>")
@default_view(forum, PostAttachment, "attachment_id", kw_func=attachment_kw_view_func)
def attachment_download(thread_id, post_id, attachment_id):
    thread = Thread.query.get(thread_id)
    post = Post.query.get(post_id)
    attachment = PostAttachment.query.get(attachment_id)

    if (
        not (thread and post and attachment)
        or post.thread is not thread
        or attachment.post is not post
    ):
        raise NotFound

    # Don't fail too hard when attachment is not available.
    try:
        content = attachment.content or ""
    except AttributeError:
        return ""

    response = make_response(content)
    response.headers["content-length"] = attachment.content_length
    response.headers["content-type"] = attachment.content_type
    filename = quote(attachment.name.encode("utf8"))
    content_disposition = f'attachment;filename="{filename}"'
    response.headers["content-disposition"] = content_disposition
    return response



from __future__ import annotations

from typing import Any

from flask import g
from flask import url_for as url_for_orig
from flask.blueprints import BlueprintSetupState
from flask_babel import lazy_gettext as _l
from flask_login import current_user

from abilian.web.action import Action, Endpoint, actions
from abilian.web.nav import NavItem

__all__ = ["register_actions"]


def url_for(endpoint: str, **kw: Any) -> str:
    return url_for_orig(endpoint, community_id=g.community.slug, **kw)


class CommunityEndpoint(Endpoint):
    def get_kwargs(self) -> dict[str, str]:
        kwargs = super().get_kwargs()
        kwargs["community_id"] = g.community.slug
        return kwargs


class CommunityTabAction(Action):
    Endpoint = CommunityEndpoint

    def url(self, context: dict[str, Any] | None = None) -> CommunityEndpoint | str:
        if self._url:
            return super().url()

        endpoint = self.endpoint
        if endpoint:
            return endpoint
        else:
            return url_for(f"{self.name}.index")

    def is_current(self) -> bool:
        return g.current_tab == self.name


_actions = (
    # Navigation
    NavItem(
        "section",
        "communities",
        title=_l("Communities"),
        url=lambda context: url_for_orig("communities.index"),
        condition=lambda ctx: current_user.is_authenticated,
    ),
    # Tabs
    CommunityTabAction("communities:tabs", "wall", _l("Activities")),
    CommunityTabAction(
        "communities:tabs",
        "documents",
        _l("Documents"),
        condition=lambda ctx: g.community.has_documents,
    ),
    CommunityTabAction(
        "communities:tabs",
        "wiki",
        _l("Wiki"),
        condition=lambda ctx: g.community.has_wiki,
    ),
    CommunityTabAction(
        "communities:tabs",
        "forum",
        _l("Conversations"),
        condition=lambda ctx: g.community.has_forum,
    ),
    CommunityTabAction(
        "communities:tabs",
        "calendar",
        _l("Calendar"),
        condition=lambda ctx: g.community.has_calendar,
    ),
    CommunityTabAction(
        "communities:tabs", "members", _l("Members"), endpoint="communities.members"
    ),
    CommunityTabAction(
        "communities:tabs",
        "settings",
        _l("Settings"),
        icon="cog",
        condition=lambda ctx: current_user.has_role("admin"),
        endpoint="communities.settings",
    ),
)


def register_actions(state: BlueprintSetupState) -> None:
    if not actions.installed(state.app):
        return
    with state.app.app_context():
        actions.register(*_actions)



from __future__ import annotations

import contextlib
import shutil
import typing
import weakref
from pathlib import Path
from typing import IO, Any
from uuid import UUID, uuid1

import sqlalchemy as sa
from flask import g

# import sqlalchemy.event
from sqlalchemy.engine.base import Connection
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm.session import Session, SessionTransaction
from sqlalchemy.orm.unitofwork import UOWTransaction

from abilian.core.extensions import db
from abilian.core.models import Model
from abilian.core.models.blob import Blob
from abilian.services import Service, ServiceState

if typing.TYPE_CHECKING:
    from abilian.app import Application

_NULL_MARK = object()


def _assert_uuid(uuid: Any):
    if not isinstance(uuid, UUID):
        raise TypeError("Not an uuid.UUID instance", uuid)


class BlobStoreServiceState(ServiceState):
    #: :class:`Path` path to application repository
    path: Path | None = None


class BlobStoreService(Service):
    """Service for storage of binary objects referenced in database."""

    name = "blob_store"
    AppStateClass = BlobStoreServiceState

    def init_app(self, app: Application):
        super().init_app(app)

        path = app.data_dir / "files"
        if not path.exists():
            path.mkdir(mode=0o775, parents=True)

        with app.app_context():
            self.app_state.path = path.resolve()

    # data management: paths and accessors
    def rel_path(self, uuid: UUID) -> Path:
        """Contruct relative path from blob store top directory to the file
        named after this uuid.

        :param:uuid: :class:`UUID` instance
        """
        _assert_uuid(uuid)

        filename = str(uuid)
        return Path(filename[0:2], filename[2:4], filename)

    def abs_path(self, uuid: UUID) -> Path:
        """Return absolute :class:`Path` object for given uuid.

        :param:uuid: :class:`UUID` instance
        """
        _assert_uuid(uuid)

        top = self.app_state.path
        rel_path = self.rel_path(uuid)
        dest = top / rel_path
        assert top in dest.parents
        return dest

    def get(self, uuid: UUID, default: Path | None = None) -> Path | None:
        """Return absolute :class:`Path` object for given uuid, if this uuid
        exists in blob store, or `default` if it doesn't.

        :param:uuid: :class:`UUID` instance
        """
        _assert_uuid(uuid)

        path = self.abs_path(uuid)
        if not path.exists():
            return default
        return path

    def set(self, uuid: UUID, content: Any, encoding: str | None = "utf-8"):
        """Store binary content with uuid as key.

        :param:uuid: :class:`UUID` instance
        :param:content: string, bytes, or any object with a `read()` method
        :param:encoding: encoding to use when content is Unicode
        """
        _assert_uuid(uuid)

        dest = self.abs_path(uuid)
        if not dest.parent.exists():
            dest.parent.mkdir(0o775, parents=True)

        if hasattr(content, "read"):
            content = content.read()

        mode = "tw"
        if not isinstance(content, str):
            mode = "bw"
            encoding = None

        with dest.open(mode, encoding=encoding) as file:
            file.write(content)

    def delete(self, uuid: UUID):
        """Delete file with given uuid.

        :param:uuid: :class:`UUID` instance
        :raises:KeyError if file does not exists
        """
        _assert_uuid(uuid)

        dest = self.abs_path(uuid)
        if not dest.exists():
            raise KeyError("No file can be found for this uuid", uuid)

        dest.unlink()

    def __getitem__(self, uuid: UUID) -> Path:
        _assert_uuid(uuid)

        value = self.get(uuid)
        if value is None:
            raise KeyError("No file can be found for this uuid", uuid)
        return value

    def __setitem__(self, uuid: UUID, content: Any):
        _assert_uuid(uuid)

        self.set(uuid, content)

    def __delitem__(self, uuid: UUID):
        _assert_uuid(uuid)

        self.delete(uuid)


blob_store = BlobStoreService()


class SessionBlobStoreState(ServiceState):
    path: Path

    @staticmethod
    def actual_session(session: Session | scoped_session) -> Session:
        if isinstance(session, scoped_session):
            return session()
        return session

    def transaction_key(self, session: Session | scoped_session) -> str:
        session_id = id(self.actual_session(session))
        return f"blob_session_{session_id}"

    # transaction <-> db session accessors
    def get_transaction(
        self, session: Session | scoped_session
    ) -> BlobStoreTransaction | None:
        session = self.actual_session(session)
        key = self.transaction_key(session)
        default = weakref.ref(session), None
        session_reference, transaction = g.get(key, default)
        found_session = session_reference()

        if found_session is None or found_session is not session:
            # old session with same id, maybe not yet garbage collected
            transaction = None
        return transaction

    def set_transaction(
        self,
        session: Session | scoped_session,
        transaction: BlobStoreTransaction | None,
    ) -> None:
        session = self.actual_session(session)
        key = self.transaction_key(session)
        value = (weakref.ref(session), transaction)
        setattr(g, key, value)

    def create_transaction(self, session: Session) -> None:
        if not self.running:
            return

        parent_transaction = self.get_transaction(session)
        transaction = BlobStoreTransaction(self.path, parent_transaction)
        self.set_transaction(session, transaction)

    def end_transaction(self, session: Session) -> None:
        if not self.running:
            return

        transaction = self.get_transaction(session)
        if transaction is not None:
            if not transaction.cleared:
                # root and nested transactions emit "commit", but
                # subtransactions don't
                transaction.commit()
            self.set_transaction(session, transaction._parent)

    def begin(self, session: Session):
        if not self.running:
            return

        transaction = self.get_transaction(session)
        if transaction is None:
            # FIXME: return or create a new one?
            return

        transaction.begin()

    def commit(self, session: Session):
        if not self.running:
            return

        transaction = self.get_transaction(session)
        if transaction is None:
            return

        transaction.commit()

    def flush(self, session: Session):
        # def flush(self, session: Session, flush_context: UOWTransaction):
        # when sqlalchemy is flushing it is done in a sub-transaction,
        # not the root one. So when calling our 'commit' from here
        # we are not in our root transaction, so changes will not be
        # written to repository.
        self.commit(session)

    def rollback(self, session: Session):
        if not self.running:
            return

        transaction = self.get_transaction(session)
        if transaction is None:
            return

        transaction.rollback()


class SessionBlobStoreService(Service):
    """A blob store service that is session aware, i.e content is actually
    written or delete at commit time.

    All content is stored using the main :class:`BlobStoreService`.
    """

    name = "session_blob_store"
    AppStateClass = SessionBlobStoreState

    def __init__(self, *args, **kwargs):
        self.__listening = False
        super().__init__(*args, **kwargs)

    def init_app(self, app: Application):
        super().init_app(app)

        path = Path(app.instance_path, "tmp", "files_transactions")
        if not path.exists():
            path.mkdir(0o775, parents=True)

        with app.app_context():
            self.app_state.path = path.resolve()

        if not self.__listening:
            self.start_listening()

    def start_listening(self):
        self.__listening = True
        listen = sa.event.listen
        listen(Session, "after_transaction_create", self.create_transaction)
        listen(Session, "after_transaction_end", self.end_transaction)
        listen(Session, "after_begin", self.begin)
        listen(Session, "after_commit", self.commit)
        listen(Session, "after_flush", self.flush)
        listen(Session, "after_rollback", self.rollback)
        # appcontext_tearing_down.connect(self.clear_transaction, app)

    def _session_for(self, model_or_session: Model | Session) -> Session:
        """Return session instance for object parameter.

        - If parameter is a session instance, it is returned as is.

        - If parameter is a registered model instance, its session will be used.

        - If parameter is a detached model instance, or None, application scoped
          session will be used (db.session()).

        - If parameter is a scoped_session instance, a new session will be
          instaniated.
        """
        if model_or_session is None:
            return db.session

        if isinstance(model_or_session, Session):
            return model_or_session

        if isinstance(model_or_session, sa.orm.scoped_session):
            return model_or_session()

        session = sa.orm.object_session(model_or_session)

        if session is None:
            return db.session

        return session

    # Blob store interface
    def get(
        self, session: Session | Blob, uuid: UUID, default: Path | None = None
    ) -> Path | None:
        # assert isinstance(session, Session)
        _assert_uuid(uuid)

        session = self._session_for(session)
        transaction = self.app_state.get_transaction(session)
        if transaction is None:
            return default
        try:
            val = transaction.get(uuid)
        except KeyError:
            return default

        if val is _NULL_MARK:
            val = blob_store.get(uuid, default)

        return val

    def set(
        self,
        session: Session | Blob,
        uuid: UUID,
        content: IO | bytes | str,
        encoding: str = "utf-8",
    ):
        _assert_uuid(uuid)
        session = self._session_for(session)
        transaction = self.app_state.get_transaction(session)
        if transaction is None:
            raise RuntimeError("transaction is None in blob store service")
        transaction.set(uuid, content, encoding)

    def delete(self, session: Session | Blob, uuid: UUID):
        _assert_uuid(uuid)

        session = self._session_for(session)
        transaction = self.app_state.get_transaction(session)
        if self.get(session, uuid) is not None:
            transaction.delete(uuid)

    # session event handlers
    @Service.if_running
    def create_transaction(
        self, session: Session, _transaction: SessionTransaction
    ) -> None:
        self.app_state.create_transaction(session)

    @Service.if_running
    def end_transaction(
        self, session: Session, _transaction: SessionTransaction
    ) -> None:
        self.app_state.end_transaction(session)

    @Service.if_running
    def begin(
        self,
        session: Session,
        _transaction: SessionTransaction,
        _connection: Connection,
    ) -> None:
        self.app_state.begin(session)

    @Service.if_running
    def commit(self, session: Session) -> None:
        self.app_state.commit(session)

    @Service.if_running
    def flush(self, session: Session, _flush_context: UOWTransaction) -> None:
        self.app_state.flush(session)

    @Service.if_running
    def rollback(self, session: Session) -> None:
        self.app_state.rollback(session)


session_blob_store = SessionBlobStoreService()


class BlobStoreTransaction:
    """Temporary local storage of file content during a transaction on
    a Blob (and related child blobs).

    Transaction shall end by either a commit or a rollback.

    Intermediate blob storage lies in self.tmp_folder and is cleared
    at end of transaction.

    BlobStoreTransaction is used by SessionBlobStoreService and the
    commit() method triggers final storage through the local "blob_store"
    service.

    Use the set() method to add actual content to the transaction and get()
    to retrieve the path of the underlying content.
    """

    def __init__(self, root_path: Path, parent: BlobStoreTransaction | None = None):
        self.tmp_folder = root_path / str(uuid1())
        # if parent is not None and parent.cleared:
        #   parent = None

        self._parent = parent
        self._deleted: set[UUID] = set()
        self._set: set[UUID] = set()
        self._cleared: bool = False

    @property
    def cleared(self) -> bool:
        return self._cleared

    def __del__(self):
        if not self._cleared:
            self._clear()

    def _clear(self):
        if self._cleared:
            return

        # make sure transaction is not usable anymore
        if self.tmp_folder.exists():
            shutil.rmtree(self.tmp_folder)

        del self.tmp_folder
        del self._deleted
        del self._set
        self._cleared = True

    def begin(self):
        if not self.tmp_folder.exists():
            self.tmp_folder.mkdir(0o700)

    def rollback(self):
        self._clear()

    def commit(self):
        """Merge modified objects into parent transaction.

        Once commited a transaction object is not usable anymore

        :param:session: current sqlalchemy Session
        """
        if self._cleared:
            return

        if self._parent:
            # nested transaction
            self._commit_parent()
        else:
            self._commit_blob_store()
        self._clear()

    def _commit_blob_store(self):
        if self._cleared:
            return
        assert self._parent is None

        for uuid in self._deleted:
            with contextlib.suppress(KeyError):
                blob_store.delete(uuid)

        for uuid in self._set:
            blob_store.set(uuid, self.uuid_path(uuid).open("rb"))

    def uuid_path(self, uuid: UUID) -> Path:
        return self.tmp_folder / str(uuid)

    def _commit_parent(self):
        parent = self._parent
        assert parent
        parent._deleted |= self._deleted
        parent._deleted -= self._set

        parent._set |= self._set
        parent._set -= self._deleted

        if self._set:
            parent.begin()  # ensure p.path exists

        for uuid in self._set:
            self.uuid_path(uuid).replace(parent.uuid_path(uuid))

    def _add_to(self, uuid: UUID, dest: set[UUID], other: set[UUID]):
        """Add `item` to `dest` set, ensuring `item` is not present in `other`
        set."""
        _assert_uuid(uuid)
        other.discard(uuid)
        dest.add(uuid)

    def delete(self, uuid: UUID):
        self._add_to(uuid, self._deleted, self._set)

    def set(
        self,
        uuid: UUID,
        content: IO | bytes | str,
        encoding: str | None = "utf-8",
    ):
        self.begin()
        self._add_to(uuid, self._set, self._deleted)

        if hasattr(content, "read"):
            content = content.read()

        if isinstance(content, bytes):
            mode = "wb"
            encoding = None
        else:
            mode = "wt"
        dest = self.uuid_path(uuid)
        with dest.open(mode, encoding=encoding) as file:
            file.write(content)

    def get(self, uuid: UUID) -> Any:
        if uuid in self._deleted:
            raise KeyError(uuid)

        if uuid in self._set:
            path = self.uuid_path(uuid)
            assert path.exists()
            return path

        if self._parent:
            return self._parent.get(uuid)

        return _NULL_MARK



"""Blueprint for views of dynamic images."""

from __future__ import annotations

import colorsys
import hashlib
from importlib import resources as rso
from pathlib import Path
from typing import Any

import sqlalchemy as sa
import sqlalchemy.orm
from flask import Blueprint, make_response, render_template, request
from werkzeug.exceptions import BadRequest, NotFound

from abilian.core.models.blob import Blob
from abilian.core.models.subjects import User
from abilian.services.image import CROP, RESIZE_MODES, get_format, get_size, resize
from abilian.web.util import url_for

from .files import BaseFileDownload

images_bp = Blueprint("images", __name__, url_prefix="/images")
route = images_bp.route

DEFAULT_AVATAR = rso.files("abilian.web") / "resources" / "img" / "avatar-default.png"
DEFAULT_AVATAR_MD5 = hashlib.md5(DEFAULT_AVATAR.read_bytes()).hexdigest()  # noqa: S324


class BaseImageView(BaseFileDownload):
    max_size = None

    def __init__(self, max_size=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Override class default value only if arg is specified in constructor.
        # This allows subclasses to easily override theses defaults.
        if max_size is not None:
            self.max_size = max_size

    def prepare_args(self, args, kwargs):
        args, kwargs = super().prepare_args(args, kwargs)
        size = request.args.get("s", 0)
        try:
            size = int(size)
        except ValueError as e:
            raise BadRequest(f'Invalid value for "s": {size:d}. Not an integer.') from e

        if self.max_size is not None and size > self.max_size:
            raise BadRequest(f"Size too large: {size:d} (max: {self.max_size:d})")

        kwargs["size"] = size

        resize_mode = request.args.get("m", CROP)
        if resize_mode not in RESIZE_MODES:
            resize_mode = CROP

        kwargs["mode"] = resize_mode
        return args, kwargs

    def make_response(self, image, size, mode, filename=None, *args, **kwargs):
        """
        :param image: image as bytes
        :param size: requested maximum width/height size
        :param mode: one of 'scale', 'fit' or 'crop'
        :param filename: filename
        """
        try:
            fmt = get_format(image)
        except OSError as e:
            # not a known image file
            raise NotFound from e

        self.content_type = "image/png" if fmt == "PNG" else "image/jpeg"
        ext = f".{fmt.lower()!s}"

        if not filename:
            filename = "image"
        if not filename.lower().endswith(ext):
            filename += ext
        self.filename = filename

        if size:
            image = resize(image, size, size, mode=mode)
            if mode == CROP:
                assert get_size(image) == (size, size)
        else:
            image = image.read()

        return make_response(image)

    def get_filename(self, *args, **kwargs):
        return self.filename

    def get_content_type(self, *args, **kwargs):
        return self.content_type


class StaticImageView(BaseImageView):
    """View for static assets not served by static directory.

    Useful for default avatars for example.
    """

    expire_vary_arg = "md5"

    def __init__(self, image, *args, **kwargs):
        """
        :param image: path to image file
        """
        super().__init__(*args, **kwargs)
        self.image_path = Path(image)
        if not self.image_path.exists():
            p = str(self.image_path)
            raise ValueError(f"Invalid image path: {p!r}")

    def prepare_args(self, args, kwargs):
        kwargs["image"] = self.image_path.open("rb")
        kwargs["filename"] = self.image_path.name
        return super().prepare_args(args, kwargs)


class BlobView(BaseImageView):
    """Default :attr:`expire_vary_arg` to `"md5"`.

    :attr:`set_expire` is set to `False` by default.
    """

    expire_vary_arg = "md5"
    id_arg = "object_id"

    def __init__(self, id_arg=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if id_arg is not None:
            self.id_arg = id_arg

    def prepare_args(self, args, kwargs):
        args, kwargs = super().prepare_args(args, kwargs)
        blob_id = kwargs.get(self.id_arg)
        try:
            blob_id = int(blob_id)
        except ValueError as e:
            raise BadRequest(f"Invalid blob id: {blob_id!r}") from e

        blob = Blob.query.get(blob_id)
        if not blob:
            raise NotFound

        meta = blob.meta
        filename = meta.get("filename", meta.get("md5", str(blob.uuid)))
        kwargs["filename"] = filename
        kwargs["image"] = blob.file.open("rb")
        return args, kwargs


blob_image = BlobView.as_view("blob_image")
route("/files/<int:object_id>")(blob_image)


class UserMugshot(BaseImageView):
    expire_vary_arg = "md5"

    def prepare_args(self, args, kwargs):
        args, kwargs = super().prepare_args(args, kwargs)

        user_id = kwargs["user_id"]
        user = User.query.options(sa.orm.undefer(User.photo)).get(user_id)

        if user is None:
            raise NotFound

        kwargs["user"] = user
        kwargs["image"] = user.photo
        return args, kwargs

    def make_response(self, user, image, size, *args, **kwargs):
        if image:
            # user has set a photo
            return super().make_response(image, size, *args, **kwargs)

        # render svg avatar
        if user.last_name:
            letter = user.last_name[0]
        elif user.first_name:
            letter = user.first_name[0]
        else:
            letter = "?"
        letter = letter.upper()
        # generate bg color, pastel: sat=65% in hsl color space
        id_hash = hash((user.name + user.email).encode("utf-8"))
        hue = id_hash % 10
        hue = (hue * 36) / 360.0  # 10 colors: 360 / 10
        color = colorsys.hsv_to_rgb(hue, 0.65, 1.0)
        color = [int(x * 255) for x in color]
        color = f"rgb({color[0]}, {color[1]}, {color[2]})"
        svg = render_template(
            "default/avatar.svg", color=color, letter=letter, size=size
        )
        response = make_response(svg)
        self.content_type = "image/svg+xml"
        self.filename = f"avatar-{id_hash}.svg"
        return response


user_photo = UserMugshot.as_view("user_photo", set_expire=True, max_size=500)
route("/users/<int:user_id>")(user_photo)
route("/users/default")(
    StaticImageView.as_view("user_default", set_expire=True, image=DEFAULT_AVATAR)
)


def user_url_args(user: User, size: int) -> tuple[str, dict[str, Any]]:
    endpoint = "images.user_default"
    kwargs = {"s": size, "md5": DEFAULT_AVATAR_MD5}

    if not user.is_anonymous:
        endpoint = "images.user_photo"
        kwargs["user_id"] = user.id
        content = user.photo if user.photo else (user.name + user.email).encode("utf-8")
        kwargs["md5"] = hashlib.md5(content).hexdigest()  # noqa: S324

    return endpoint, kwargs


def user_photo_url(user, size):
    """Return url to use for this user."""
    endpoint, kwargs = user_url_args(user, size)
    return url_for(endpoint, **kwargs)



"""Base Flask application class, used by tests or to be extended in real
applications."""

from __future__ import annotations

import errno
import importlib

# Temps monkey patches
import os
import sys
import warnings
from collections.abc import Callable, Collection
from functools import cached_property, partial
from itertools import chain, count
from pathlib import Path
from typing import Any

import defusedxml
import jinja2
import sqlalchemy as sa
import sqlalchemy.exc
from flask import (
    Blueprint,
    Flask,
    abort,
    appcontext_pushed,
    g,
    request,
    request_started,
)
from flask.config import Config, ConfigAttribute
from flask_migrate import Migrate
from flask_tailwind import Tailwind
from flask_talisman import DEFAULT_CSP_POLICY, Talisman
from loguru import logger

import abilian.core.util
import abilian.i18n
from abilian.config import default_config
from abilian.core import extensions, signals
from abilian.services import (
    Service,
    activity_service,
    antivirus,
    audit_service,
    auth_service,
    blob_store,
    conversion_service,
    index_service,
    preferences_service,
    security_service,
    session_blob_store,
    settings_service,
    vocabularies_service,
)
from abilian.services.security import Anonymous
from abilian.services.security.models import Role
from abilian.web import csrf
from abilian.web.access_blueprint import allow_access_for_roles
from abilian.web.action import actions
from abilian.web.admin import Admin
from abilian.web.assets import AssetManagerMixin
from abilian.web.errors import ErrorManagerMixin
from abilian.web.jinja import JinjaManagerMixin
from abilian.web.nav import BreadcrumbItem
from abilian.web.util import send_file_from_directory
from abilian.web.views import Registry as ViewRegistry

defusedxml.defuse_stdlib()

db = extensions.db
__all__ = ["Application", "ServiceManager", "create_app"]

# Silence those warnings for now.
warnings.simplefilter("ignore", category=sa.exc.SAWarning)


class ServiceManager:
    """Mixin that provides lifecycle (register/start/stop) support for
    services."""

    services: dict[str, Service]

    def __init__(self):
        self.services = {}

    def start_services(self):
        for svc in self.services.values():
            svc.start()

    def stop_services(self):
        for svc in self.services.values():
            svc.stop()


class PluginManager:
    """Mixin that provides support for loading plugins."""

    config: Config

    #: Custom apps may want to always load some plugins: list them here.
    APP_PLUGINS = [
        "abilian.web.search",
        "abilian.web.tags",
        "abilian.web.comments",
        "abilian.web.uploads",
        "abilian.web.attachments",
    ]

    def register_plugin(self, name: str):
        """Load and register a plugin given its package name."""
        logger.info("Registering plugin: {name}", name=name)
        module = importlib.import_module(name)
        module.register_plugin(self)  # type: ignore

    def register_plugins(self):
        """Load plugins listed in config variable 'PLUGINS'."""
        registered = set()
        for plugin_fqdn in chain(self.APP_PLUGINS, self.config["PLUGINS"]):
            if plugin_fqdn not in registered:
                self.register_plugin(plugin_fqdn)
                registered.add(plugin_fqdn)


class Application(
    ServiceManager,
    PluginManager,
    AssetManagerMixin,
    ErrorManagerMixin,
    JinjaManagerMixin,
    Flask,
):
    """Base application class.

    Extend it in your own app.
    """

    default_config = default_config

    #: If True all views will require by default an authenticated user, unless
    #: Anonymous role is authorized. Static assets are always public.
    private_site = ConfigAttribute("PRIVATE_SITE")

    #: instance of :class:`.web.views.registry.Registry`.
    default_view: ViewRegistry

    #: json serializable dict to land in Javascript under Abilian.api
    js_api: dict[str, Any]

    def __init__(self, name: Any | None = None, *args: Any, **kwargs: Any):
        name = name or __name__

        Flask.__init__(self, name, *args, **kwargs)

        ServiceManager.__init__(self)
        PluginManager.__init__(self)
        JinjaManagerMixin.__init__(self)

        self.default_view = ViewRegistry()
        self.js_api = {}

    def setup(self, config: type | None):
        self.configure(config)

        # At this point we have loaded all external config files:
        # SQLALCHEMY_DATABASE_URI is definitively fixed (it cannot be defined in
        # database AFAICT), and LOGGING_FILE cannot be set in DB settings.
        self.setup_logging()

        appcontext_pushed.connect(self.install_id_generator)

        if not self.testing:
            self.init_sentry()

        # time to load config bits from database: 'settings'
        # First init required stuff: db to make queries, and settings service
        extensions.db.init_app(self)
        settings_service.init_app(self)

        self.register_jinja_loaders(jinja2.PackageLoader("abilian.web"))
        self.init_assets()
        self.install_default_handlers()

        with self.app_context():
            self.init_extensions()
            self.register_plugins()
            self.add_access_controller(
                "static", allow_access_for_roles(Anonymous), endpoint=True
            )
            # debugtoolbar: this is needed to have it when not authenticated
            # on a private site. We cannot do this in init_debug_toolbar,
            # since auth service is not yet installed.
            self.add_access_controller(
                "debugtoolbar", allow_access_for_roles(Anonymous)
            )
            self.add_access_controller(
                "_debug_toolbar.static",
                allow_access_for_roles(Anonymous),
                endpoint=True,
            )

        # TODO: maybe reenable later
        # self.maybe_register_setup_wizard()

        self._finalize_assets_setup()

        # At this point all models should have been imported: time to configure
        # mappers. Normally Sqlalchemy does it when needed but mappers may be
        # configured inside sa.orm.class_mapper() which hides a
        # misconfiguration: if a mapper is misconfigured its exception is
        # swallowed by class_mapper(model) results in this laconic
        # (and misleading) message: "model is not mapped"
        sa.orm.configure_mappers()

        signals.components_registered.send(self)

        request_started.connect(self.setup_nav_and_breadcrumbs)

        if not self.testing:
            signals.register_js_api.send(self)

        # Initialize Abilian core services.
        # Must come after all entity classes have been declared.
        # Inherited from ServiceManager. Will need some configuration love
        # later.
        if not self.testing:
            with self.app_context():
                self.start_services()

        setup(self)

    def setup_nav_and_breadcrumbs(self, app: Flask):
        """Listener for `request_started` event.

        If you want to customize first items of breadcrumbs, override
        :meth:`init_breadcrumbs`
        """
        g.nav = {"active": None}  # active section
        g.breadcrumb = []
        self.init_breadcrumbs()

    def init_breadcrumbs(self):
        """Insert the first element in breadcrumbs.

        This happens during `request_started` event, which is triggered
        before any url_value_preprocessor and `before_request` handlers.
        """
        g.breadcrumb.append(BreadcrumbItem(icon="home", url=f"/{request.script_root}"))

    # TODO: remove
    def install_id_generator(self, sender: Flask, **kwargs: Any):
        g.id_generator = count(start=1)

    def configure(self, config: type | None):
        if config:
            self.config.from_object(config)
        else:
            self.config.from_prefixed_env()

        # Setup babel config
        languages = self.config["BABEL_ACCEPT_LANGUAGES"]
        languages = tuple(
            lang for lang in languages if lang in abilian.i18n.VALID_LANGUAGES_CODE
        )
        self.config["BABEL_ACCEPT_LANGUAGES"] = languages

        # This needs to be done dynamically
        if not self.config.get("SESSION_COOKIE_NAME"):
            self.config["SESSION_COOKIE_NAME"] = f"{self.name}-session"

        if not self.config.get("FAVICO_URL"):
            self.config["FAVICO_URL"] = self.config.get("LOGO_URL")

        if not self.debug and self.config["SECRET_KEY"] == "CHANGEME":  # noqa: S105
            logger.error("You must change the default secret config ('SECRET_KEY')")
            sys.exit()

    def check_instance_folder(self, create=False):
        """Verify instance folder exists, is a directory, and has necessary
        permissions.

        :param:create: if `True`, creates directory hierarchy

        :raises: OSError with relevant errno if something is wrong.
        """
        path = Path(self.instance_path)
        err = None
        eno = 0

        if not path.exists():
            if create:
                logger.info("Create instance folder: {path}", path=str(path))
                path.mkdir(0o775, parents=True)
            else:
                err = "Instance folder does not exists"
                eno = errno.ENOENT
        elif not path.is_dir():
            err = "Instance folder is not a directory"
            eno = errno.ENOTDIR
        elif not os.access(str(path), os.R_OK | os.W_OK | os.X_OK):
            err = 'Require "rwx" access rights, please verify permissions'
            eno = errno.EPERM

        if err:
            raise OSError(eno, err, str(path))

    @cached_property
    def data_dir(self) -> Path:
        path = Path(self.instance_path, "data")
        if not path.exists():
            path.mkdir(0o775, parents=True)

        return path

    def _init_babel(self):
        # Babel (for i18n)
        babel = abilian.i18n.babel
        # Temporary (?) workaround
        babel.locale_selector_func = None
        babel.timezone_selector_func = None

        babel.init_app(self)

        babel.add_translations("wtforms", translations_dir="locale", domain="wtforms")
        babel.add_translations("abilian")

    def init_extensions(self):
        """Initialize flask extensions, helpers and services."""

        self._init_babel()

        extensions.redis.init_app(self)
        extensions.mail.init_app(self)
        extensions.deferred_js.init_app(self)
        extensions.upstream_info.extension.init_app(self)
        actions.init_app(self)

        # auth_service installs a `before_request` handler (actually it's
        # flask-login). We want to authenticate user ASAP, so that sentry and
        # logs can report which user encountered any error happening later,
        # in particular in a before_request handler (like csrf validator)
        auth_service.init_app(self)

        # webassets
        self.setup_asset_extension()
        self.register_base_assets()

        # Flask-Migrate
        Migrate(self, db)

        # CSRF by default
        if self.config.get("WTF_CSRF_ENABLED"):
            # extensions.csrf IS original wtf_csrf
            # abilian_csrf is
            # from .csrf import abilian_csrf
            # from .csrf import wtf_csrf as csrf

            extensions.csrf.init_app(self)  # double initialization
            self.extensions["csrf"] = extensions.csrf
            extensions.abilian_csrf.init_app(self)

        self.register_blueprint(csrf.csrf_blueprint)

        # images blueprint
        from .web.views.images import images_bp

        self.register_blueprint(images_bp)

        # Abilian Core services
        security_service.init_app(self)
        blob_store.init_app(self)
        session_blob_store.init_app(self)
        audit_service.init_app(self)
        index_service.init_app(self)
        activity_service.init_app(self)
        preferences_service.init_app(self)
        conversion_service.init_app(self)
        vocabularies_service.init_app(self)
        antivirus.init_app(self)

        from .web.preferences.user import UserPreferencesPanel

        preferences_service.register_panel(UserPreferencesPanel(), self)

        from .web.coreviews import users

        self.register_blueprint(users.blueprint)

        # Admin interface
        Admin().init_app(self)

        # dev helper
        if self.debug:
            # during dev, one can go to /http_error/403 to see rendering of 403
            http_error_pages = Blueprint("http_error_pages", __name__)

            @http_error_pages.route("/<int:code>")
            def error_page(code):
                """Helper for development to show 403, 404, 500..."""
                abort(code)

            self.register_blueprint(http_error_pages, url_prefix="/http_error")

    def add_url_rule_with_role(
        self,
        rule: str,
        endpoint: str,
        view_func: Callable,
        roles: Collection[Role] = (),
        **options: Any,
    ):
        """See :meth:`Flask.add_url_rule`.

        If `roles` parameter is present, it must be a
        :class:`abilian.service.security.models.Role` instance, or a list of
        Role instances.
        """
        self.add_url_rule(rule, endpoint, view_func, **options)

        if roles:
            self.add_access_controller(
                endpoint, allow_access_for_roles(roles), endpoint=True
            )

    def add_access_controller(self, name: str, func: Callable, endpoint: bool = False):
        """Add an access controller.

        If `name` is None it is added at application level, else if is
        considered as a blueprint name. If `endpoint` is True then it is
        considered as an endpoint.
        """
        auth_state = self.extensions[auth_service.name]

        if endpoint:
            if not isinstance(name, str):
                msg = f"{name!r} is not a valid endpoint name"
                raise ValueError(msg)

            auth_state.add_endpoint_access_controller(name, func)
        else:
            auth_state.add_bp_access_controller(name, func)

    def add_static_url(
        self, url_path: str, directory: str, endpoint: str, roles: Collection[Role] = ()
    ):
        """Add a new url rule for static files.

        :param url_path: subpath from application static url path. No heading
            or trailing slash.
        :param directory: directory to serve content from.
        :param endpoint: flask endpoint name for this url rule.

        Example::

           app.add_static_url('myplugin',
                              '/path/to/myplugin/resources',
                              endpoint='myplugin_static')

        With default setup it will serve content from directory
        `/path/to/myplugin/resources` from url `http://.../static/myplugin`
        """
        url_path = f"{self.static_url_path}/{url_path}/<path:filename>"
        self.add_url_rule_with_role(
            url_path,
            endpoint=endpoint,
            view_func=partial(send_file_from_directory, directory=directory),
            roles=roles,
        )
        self.add_access_controller(
            endpoint, allow_access_for_roles(Anonymous), endpoint=True
        )


def setup(app: Flask):
    config = app.config

    # CSP
    if not app.debug:
        csp = config.get("CONTENT_SECURITY_POLICY", DEFAULT_CSP_POLICY)
        Talisman(app, content_security_policy=csp)

    Tailwind(app)

    # Debug Toolbar
    init_debug_toolbar(app)


def init_debug_toolbar(app: Flask):
    if not app.debug or app.testing:
        return

    try:
        from flask_debugtoolbar import DebugToolbarExtension
    except ImportError:
        logger.warning("Running in debug mode but flask_debugtoolbar is not installed.")
        return

    dbt = DebugToolbarExtension()
    default_config = dbt._default_config(app)
    init_dbt = dbt.init_app

    if "DEBUG_TB_PANELS" not in app.config:
        # add our panels to default ones
        app.config["DEBUG_TB_PANELS"] = list(default_config["DEBUG_TB_PANELS"])
    init_dbt(app)
    for view_name in app.view_functions:
        if view_name.startswith("debugtoolbar."):
            extensions.csrf.exempt(app.view_functions[view_name])


def create_app(
    config: type | None = None, app_class: type = Application, **kw: Any
) -> Application:
    app = app_class(**kw)
    app.setup(config=config)

    # This is currently called from app.setup()
    # setup(app)

    return app
